'''
Date: 2024-05-06 17:40:28
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-11-23 20:21:30
Description: 
'''

from .__version__ import (__author__, __author_email__, __build__,
                          __description__, __license__, __title__, __url__,
                          __version__)